/*
 * Some code borrowed from sg-utils.
 *
 * Copyright (c) 2004 Christophe Varoqui
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include "checkers.h"

#include "../libmultipath/debug.h"
#include "../libmultipath/sg_include.h"

#define TUR_CMD_LEN 6
#define HEAVY_CHECK_COUNT       10

#ifdef CHECK_WWID
/* from linux/include/scsi/scsi.h */
#define DID_BUS_BUSY    0x02
#define DID_ERROR       0x07
#define DID_TRANSPORT_DISRUPTED 0x0e

#define MSG_TUR_UP     "HP tur checker reports path is up"
#define MSG_TUR_DOWN   "HP tur checker reports path is down"
#define MSG_TUR_GHOST  "HP tur checker reports path is in standby state"
#define EVPD            0x01
#define PAGE_83         0x83
#define INQUIRY_CMD     0x12
#define INQUIRY_CMDLEN  6
#define SCSI_INQ_BUFF_LEN 96
#else
#define MSG_TUR_UP	"tur checker reports path is up"
#define MSG_TUR_GHOST	"tur checker reports path in standby state"
#define MSG_TUR_DOWN	"tur checker reports path is down"
#endif

#ifdef CHECK_WWID
static int
do_inq(struct checker * c, char * wwid)
{
	int ret = -1;
	unsigned char inq_cmd[INQUIRY_CMDLEN] =
	{INQUIRY_CMD, EVPD, PAGE_83, 0, SCSI_INQ_BUFF_LEN, 0 };
	unsigned char sense_buffer[32];
	unsigned char resp_buffer[SCSI_INQ_BUFF_LEN];
	char *pbuff;

	int m,k;
	int retry_tur = 5;
	struct sg_io_hdr io_hdr;

retry:
	memset(resp_buffer, 0, sizeof(resp_buffer));
	memset(&io_hdr, 0, sizeof(struct sg_io_hdr));

	io_hdr.interface_id = 'S';
	io_hdr.cmd_len = sizeof(inq_cmd);
	io_hdr.mx_sb_len = sizeof(sense_buffer);
	io_hdr.dxfer_direction = -3; // Data transfer from the device.
	io_hdr.dxfer_len = sizeof(resp_buffer);
	io_hdr.dxferp = (unsigned char *)resp_buffer;
	io_hdr.cmdp = inq_cmd;
	io_hdr.sbp = sense_buffer;
	io_hdr.timeout = 60; // IOCTL timeout value.

	if (ioctl(c->fd, SG_IO, &io_hdr) < 0) {
		condlog(0, "SG_IO ioctl failed: %s", strerror(errno));
		return ret;
	}
	if (io_hdr.info & SG_INFO_OK_MASK){
		int key = 0, asc, ascq;

		if (io_hdr.host_status == DID_BUS_BUSY ||
				io_hdr.host_status == DID_ERROR ||
				io_hdr.host_status == DID_TRANSPORT_DISRUPTED) {
			if (--retry_tur)
				goto retry;
		}
		if (io_hdr.sb_len_wr > 3) {
			if (io_hdr.sbp[0] == 0x72 || io_hdr.sbp[0] == 0x73) {
				key = io_hdr.sbp[1] & 0x0f;
				asc = io_hdr.sbp[2];
				ascq = io_hdr.sbp[3];
			} else if (io_hdr.sb_len_wr > 13 &&
					((io_hdr.sbp[0] & 0x7f) == 0x70 ||
					 (io_hdr.sbp[0] & 0x7f) == 0x71)) {
				key = io_hdr.sbp[2] & 0x0f;
				asc = io_hdr.sbp[12];
				ascq = io_hdr.sbp[13];
			}
		}
		if (key == 0x6) {
			/* Unit Attention, retry */
			if (--retry_tur)
				goto retry;
		}
		return ret;
	}

	pbuff = (char *) resp_buffer;

	wwid[0] = '3';
	for (m = 8, k = 1; m < 11; ++m, k+=2)
		sprintf(&wwid[k], "%02x", (unsigned int)pbuff[m] & 0xff);
	for (m = 11; m < 24; ++m, k+=2)
		sprintf(&wwid[k], "%02x", (unsigned int)pbuff[m] & 0xff);

	return (ret = 0);
}
#endif


/* from linux/include/scsi/scsi.h */
#define DID_BUS_BUSY    0x02
#define DID_ERROR       0x07
#define DID_TRANSPORT_DISRUPTED 0x0e

#ifdef CHECK_WWID
struct hp_tur_checker_context {
#else
struct tur_checker_context {
#endif
	void * dummy;
};

#ifdef CHECK_WWID
int hp_tur_init (struct checker * c)
{
	memset(c->wwid, 0, WWID_SIZE);
#else
int tur_init (struct checker * c)
{
#endif
	return 0;
}

#ifdef CHECK_WWID
void hp_tur_free (struct checker * c)
#else
void tur_free (struct checker * c)
#endif
{
	return;
}

extern int
#ifdef CHECK_WWID
hp_tur (struct checker * c)
#else
tur (struct checker * c)
#endif
{
	struct sg_io_hdr io_hdr;
	unsigned char turCmdBlk[TUR_CMD_LEN] = { 0x00, 0, 0, 0, 0, 0 };
	unsigned char sense_buffer[32];
	int retry_tur = 5;
#ifdef CHECK_WWID
	char wwid[WWID_SIZE];
#endif

retry:
	memset(&io_hdr, 0, sizeof (struct sg_io_hdr));
	io_hdr.interface_id = 'S';
	io_hdr.cmd_len = sizeof (turCmdBlk);
	io_hdr.mx_sb_len = sizeof (sense_buffer);
	io_hdr.dxfer_direction = SG_DXFER_NONE;
	io_hdr.cmdp = turCmdBlk;
	io_hdr.sbp = sense_buffer;
	io_hdr.timeout = c->timeout;
	io_hdr.pack_id = 0;
	if (ioctl(c->fd, SG_IO, &io_hdr) < 0) {
		MSG(c, MSG_TUR_DOWN);
		return PATH_DOWN;
	}
	if (io_hdr.info & SG_INFO_OK_MASK) {
		int key = 0, asc, ascq;

		if (io_hdr.host_status == DID_BUS_BUSY ||
		    io_hdr.host_status == DID_ERROR ||
		    io_hdr.host_status == DID_TRANSPORT_DISRUPTED) {
			if (--retry_tur)
				goto retry;
		}
		if (io_hdr.sb_len_wr > 3) {
			if (io_hdr.sbp[0] == 0x72 || io_hdr.sbp[0] == 0x73) {
				key = io_hdr.sbp[1] & 0x0f;
				asc = io_hdr.sbp[2];
				ascq = io_hdr.sbp[3];
			} else if (io_hdr.sb_len_wr > 13 &&
					((io_hdr.sbp[0] & 0x7f) == 0x70 ||
					 (io_hdr.sbp[0] & 0x7f) == 0x71)) {
				key = io_hdr.sbp[2] & 0x0f;
				asc = io_hdr.sbp[12];
				ascq = io_hdr.sbp[13];
			}
		}
		if (key == 0x6) {
			/* Unit Attention, retry */
			if (--retry_tur)
				goto retry;
		}
		else if( key == 0x2){
			/* Not Ready */
			
			/*Note : Other ALUA states are either UP or DOWN*/
			if( asc == 0x04 && ascq == 0x0b){
			/*LOGICAL UNIT NOT ACCESSIBLE, TARGET PORT IN STANDBY STATE*/
				MSG(c, MSG_TUR_GHOST);
				return PATH_GHOST;
			}
		}

		MSG(c, MSG_TUR_DOWN);
		return PATH_DOWN;
	}
#ifdef CHECK_WWID
	if (!do_inq(c, wwid)) {

		if(!strcmp(c->wwid, "\0")) {
			strcpy(c->wwid, wwid);
			goto up;
		}

		if (strcmp(c->wwid , wwid)) {
			condlog(0,
				"hp_tur: Lun collided. new_wwid %s old_wwid %s",
				wwid, c->wwid);
			MSG(c, MSG_TUR_DOWN);
			return PATH_DOWN;
		}
	}
up:
#endif
	MSG(c, MSG_TUR_UP);
	return PATH_UP;
}
